# Writing an HTTP Server in Cython+

[Cython+](https://www.cython.plus) is an ongoing project to extend [Cython](https://cython.org/) with [high performance GIL-free types](https://www.youtube.com/watch?v=Lb6q1hhBp-k) using static typing.

Our [next steps](https://www.cython.plus/P-CYP-Roadmap) include with writing a minimal standard library.

To drive this effort I'm writing a very basic HTTP server in Cython+, building everything I will need along the way.


## Usage

The repository looks like this:

```
- README.md
- src/
    - Makefile
    - setup.py
    - stdlib/
        - *.pxd
        - ...
    - *.pyx
    - ...
- ...
```

Each `.pyx` file in `src` is a standalone demo.

To run the demos (e.g. `helloworld.pyx`):

```
cd src
make helloworld
```


## GIL-free strings

So far Cython+ merely relied on C++'s `std::string` to represent strings. As it happens, writing our own GIL-free string implementation is one of our next steps.

As I started thinking about how to wrap the C socket API, I realised that it would go much easier with our own strings. In particular, C APIs often accept `null` as a valid value for `char *` string arguments, but there is no good way to represent `null` with `std::string`.

So I wrote a minimal string implementation by wrapping a `std::string` into a Cython+ GIL-free type. There are multiple advantages:
- Aliasing a string will not create a copy, unlike `std::string`
- We can write our own methods, e.g. to mirror Python's strings
- We can enforce immutability if we want

Implemented string methods include:
- `find`
- `substr` 	(plays the role of Python's slice operator)
- `__add__`	(string concatenation using `+`)
- `split`
- `join`

Since Cython+ GIL-free types are allocated to the heap and referenced by pointer, `null` is a valid value.

Actually at first I wanted to avoid `std::string` because they incur unneeded but unavoidable zeroing or copying overhead when interacting with C I/O APIs. This is very well explained in [the `basic_string::resize_and_overwrite` proposal for C++23](http://www.open-std.org/JTC1/SC22/WG21/docs/papers/2021/p1072r7.html). So the first version actually used `std::string_view` (C++17) with hand-crafted memory allocation using `malloc` and `free`. But that was incompatible with some C++ string libraries like the excellent [fmt](https://fmt.dev/), and that's why I went back to using `std::string` as backend.

That initial implementation remains available in branch `string`.


## Echo Server and Client

Using [Beej's Guide to Network Programming](https://beej.us/guide/bgnet/html//index.html) as a handy reference, I wrote adapters for needed parts of the POSIX socket API. As much as practical, these adapters mirror [Python's socket API](https://docs.python.org/3/library/socket.html).

These adapters include the functions:
- `getaddrinfo`
- `socket`
- `setsockopt`	(only for options with integer values, which is all I needed)
- `connect`
- `bind`
- `listen`
- `accept`
- `recv`
- `send`
- `shutdown`
- `close`

The new GIL-free strings proved very useful for this :)

Since currently Cython+ doesn't have GIL-free exceptions (also in our next steps), I take the GIL and raise Python exceptions instead when there is an error.

It's not complete, but it's enough to write a simple echo server and client demo:

```
make echoserver
```

and

```
make client
```

